import { assignDeepOmitBlank, ChartDimensionRoleType, ChartOrient, cloneDeep, getChartCategory, getChartCategory2, getEntityHierarchy, getEntityLevel, getEntityProperty, getPropertyCaption, getPropertyHierarchy, mergeOptions, omitBlank } from '@metad/ocap-core';
import format from 'date-fns/format';
import parse from 'date-fns/parse';
import { HeatmapChart } from 'echarts/charts';
import { use } from 'echarts/core';
import { isEmpty, isNil, maxBy, minBy } from 'lodash-es';
import { getCoordinateSystem, measuresToSeriesComponents, serializeSeriesComponent } from './cartesian';
import { formatMeasureNumber } from './common';
import { coordinates, gatherCoordinates } from './coordinates';
import { getEChartsTooltip } from './tooltip';
import { AxisEnum } from './types';
import { getColorLightnessVisualMap, getMeasurePaletteVisualMap } from './visualMap';
use([HeatmapChart]);
export function heatmap(data, chartAnnotation, entityType, settings, options) {
    const context = {
        data,
        entityType,
        settings,
        options,
        chartAnnotation
    };
    const type = 'heatmap';
    if (chartAnnotation.chartType.variant === 'calendar') {
        const calendar = getChartCalendar(chartAnnotation, entityType);
        return {
            ...context,
            echartsOptions: gatherCoordinates(coordinates(context, 'grid', calendarCoordinate), type, options),
            onClick: (event) => {
                return {
                    ...event,
                    event: event.event?.event,
                    filter: {
                        dimension: calendar,
                        members: [
                            event.data.member
                        ]
                    }
                };
            }
        };
    }
    const coordinateDatas = coordinates(context, 'grid', cartesianCoordinate);
    return {
        ...context,
        echartsOptions: gatherCoordinates(coordinateDatas, type, options),
    };
}
/**
 * 单个笛卡尔坐标系, 对应一个 grid 内的组件
 *
 * @param data 数据数组
 * @param chartAnnotation 图形注解
 * @param entityType 实体类型
 * @param settings 图形设置
 * @param options 图形属性
 * @returns
 */
export function cartesianCoordinate(context, data) {
    // 智能补全信息
    const index = context.chartAnnotation.dimensions.findIndex((item) => item.role === ChartDimensionRoleType.Category2);
    if (index === -1) {
        context.chartAnnotation = cloneDeep(context.chartAnnotation);
        const dimension = context.chartAnnotation.dimensions.slice(1).find((item) => !item.role);
        if (dimension) {
            dimension.role = ChartDimensionRoleType.Category2;
        }
    }
    const { chartAnnotation, entityType, settings, options } = context;
    const category = getChartCategory(chartAnnotation);
    const categoryProperty = getEntityProperty(entityType, category);
    const category2 = getChartCategory2(chartAnnotation);
    // For the moment: support only one measure
    const measure = chartAnnotation.measures?.[0];
    context.datasets = [];
    const seriesComponents = measuresToSeriesComponents(chartAnnotation.measures, data, entityType, settings);
    data = [...data];
    // TODO 是不是应该提到所有图形之前 ?
    // Fill up measure null value
    if (data[0]) {
        data[0] = { ...data[0] };
        chartAnnotation.measures.forEach((measure) => {
            if (isNil(data[0][measure.measure])) {
                data[0][measure.measure] = '-';
            }
        });
    }
    context.datasets = [
        {
            dataset: {
                source: data
            },
            seriesComponents,
            tooltip: getEChartsTooltip(options?.tooltip, categoryProperty, chartAnnotation.measures.map((measure) => ({
                measure,
                property: getEntityProperty(entityType, measure)
            })), seriesComponents, settings.locale)
        }
    ];
    const { categoryAxis, valueAxis } = getCoordinateSystem(context, data, settings.locale);
    const gridOptions = {
        type: 'Cartesian2d',
        name: '',
        grid: mergeOptions({
            top: 0,
            right: 0,
            bottom: 0,
            left: 0,
            containLabel: true
        }, options?.grid),
        [categoryAxis.orient]: categoryAxis.axis,
        [valueAxis.orient]: valueAxis.axis,
        visualMap: [],
        datasets: [],
        tooltip: [],
        legend: [],
        dataZoom: []
    };
    if (categoryAxis.dataZoom) {
        gridOptions.dataZoom.push(categoryAxis.dataZoom);
    }
    if (valueAxis.dataZoom) {
        gridOptions.dataZoom.push(valueAxis.dataZoom);
    }
    context.datasets.forEach(({ dataset, seriesComponents, tooltip }) => {
        gridOptions.datasets.push({
            dataset,
            series: seriesComponents.map((seriesComponent) => {
                const { series, visualMaps } = serializeSeriesComponent(dataset, seriesComponent, entityType, valueAxis, settings, options);
                gridOptions.visualMap.push(...visualMaps);
                if (!isNil(seriesComponent.valueAxisIndex)) {
                    series[valueAxis.orient + 'Index'] = seriesComponent.valueAxisIndex;
                }
                /**
                 * 堆积系列情况下不适用 encode, 使用的应该是 seriesLayoutBy: "row" name: "[Canada]" 与 dataset 中的行数据进行的对应
                 */
                if (isNil(series.seriesLayoutBy)) {
                    series.encode = {
                        [AxisEnum[categoryAxis.orient]]: getPropertyHierarchy(category),
                        [AxisEnum[valueAxis.orient]]: category2 ? getPropertyHierarchy(category2) : seriesComponent.measure,
                        value: chartAnnotation.measures[0].measure,
                        tooltip: seriesComponent.tooltip
                    };
                }
                return assignDeepOmitBlank(series, options?.seriesStyle, 2);
            })
        });
        // Default VisualMap
        if (isEmpty(gridOptions.visualMap)) {
            const measure = chartAnnotation.measures[0];
            const lightnessProperty = getEntityProperty(entityType, measure);
            gridOptions.visualMap.push(assignDeepOmitBlank(getColorLightnessVisualMap(measure, lightnessProperty, dataset, options?.visualMaps), measure.chartOptions?.visualMap));
        }
    });
    gridOptions.tooltip.push(assignDeepOmitBlank({
        valueFormatter: (value) => {
            return formatMeasureNumber({ measure, property: null }, value, settings.locale, measure.formatting?.shortNumber);
        },
        axisPointer: {
            type: 'cross',
            label: {
            // formatter have no effect 
            // formatter: (params) => {
            // }
            }
        }
    }, options?.tooltip, 4));
    // legend
    if (options?.legend) {
        gridOptions.legend.push(assignDeepOmitBlank({}, options.legend));
    }
    // Userdefined VisualMaps
    gridOptions.visualMap = gridOptions.visualMap.map((item, index) => {
        return {
            ...item,
            ...omitBlank(options?.visualMaps?.[index])
        };
    });
    return gridOptions;
}
export function calendarCoordinate(context, data) {
    const { chartAnnotation, entityType, settings, options } = context;
    const measure = chartAnnotation.measures[0];
    const calendar = getChartCalendar(chartAnnotation, entityType);
    const calendarHierarchy = getEntityHierarchy(entityType, calendar);
    const calendarCaption = getPropertyCaption(calendarHierarchy);
    const dateFormatter = options?.seriesStyle?.dateFormatter;
    const calendarLevelProperty = getEntityLevel(entityType, calendar);
    const levelFormatter = calendarLevelProperty.semantics?.formatter;
    function formatter(row) {
        if (!row) {
            return '-';
        }
        // 以 caption 为日期是因为 memberKey 中会有重复的年月等信息, 在格式化时不被允许
        return format(dateFormatter
            ? parse(row[calendarCaption], dateFormatter, new Date())
            : levelFormatter
                ? parse(row[calendarHierarchy.name], levelFormatter, new Date())
                : parse(row[calendarCaption], 'yyyyMMdd', new Date()), 'yyyy-MM-dd');
    }
    const seriesComponents = measuresToSeriesComponents(chartAnnotation.measures, data, entityType, settings);
    const calendarMin = minBy(data, calendarCaption);
    const calendarMax = maxBy(data, calendarCaption);
    const dataset = {
        source: data
    };
    const gridOptions = {
        type: 'Calendar',
        name: '',
        datasets: [],
        visualMap: [],
        grid: {},
        calendar: assignDeepOmitBlank({
            orient: chartAnnotation.chartType.orient === ChartOrient.vertical ? 'vertical' : 'horizontal',
            range: [formatter(calendarMin), formatter(calendarMax)]
        }, options?.calendar),
        tooltip: [],
        dataZoom: []
    };
    gridOptions.datasets = [
        {
            dataset,
            series: seriesComponents.map((seriesComponent) => {
                const { series, visualMaps } = serializeSeriesComponent(dataset, seriesComponent, entityType, null, settings, options);
                /**
                 * @todo 不使用 visualMaps 是因为其 visualMap 中有 dimensions 这个在 calendar series 情况下有问题, 还没有找到统一的解决方案
                 */
                if (seriesComponent.palette?.name) {
                    const visualMap = getMeasurePaletteVisualMap(seriesComponent.palette, seriesComponent, dataset.source);
                    gridOptions.visualMap.push(visualMap);
                }
                // series.tooltip = tooltip
                series.coordinateSystem = 'calendar';
                series.data = data.map((item) => ({
                    value: [
                        formatter(item),
                        item[measure.measure],
                    ],
                    data: item,
                    member: {
                        value: item[calendarHierarchy.name],
                        label: item[calendarCaption]
                    }
                }));
                return assignDeepOmitBlank(series, options?.seriesStyle);
            })
        }
    ];
    gridOptions.tooltip.push(assignDeepOmitBlank({
        trigger: 'item',
        valueFormatter: (value) => {
            return formatMeasureNumber({ measure, property: null }, value, settings.locale, measure.formatting?.shortNumber);
        }
    }, options?.tooltip));
    // Default VisualMap
    if (isEmpty(gridOptions.visualMap)) {
        const measure = chartAnnotation.measures[0];
        const seriesComponent = seriesComponents[0];
        // const lightnessProperty = getEntityProperty(entityType, measure)
        gridOptions.visualMap.push(assignDeepOmitBlank({
            min: seriesComponent.dataMin,
            max: seriesComponent.dataMax,
            text: [seriesComponent.property.caption || seriesComponent.property.name]
        }, measure.chartOptions?.visualMap));
    }
    // Userdefined VisualMaps
    gridOptions.visualMap = gridOptions.visualMap.map((item, index) => {
        return {
            ...item,
            ...omitBlank(options?.visualMaps?.[index])
        };
    });
    return gridOptions;
}
export function getChartCalendar(chartAnnotation, entityType) {
    return chartAnnotation.dimensions.find((item) => {
        const property = getEntityProperty(entityType, item);
        if (property.semantics?.semantic) {
            return true;
        }
        return false;
    });
}
//# sourceMappingURL=heatmap.js.map